/// The main application
module app;

import std.stdio, std.getopt, std.file, std.path, std.conv, std.string;
import parser, converter;

int main(string[] a) {
	struct Args {
		string input, output, template_ = "";
		Format fmt;
		bool dir, verbose;
	}
	Args args;
	try {
		bool empty = a.length == 1;
		auto res = getopt(
			a,
			"i|input", "Input file", &args.input,
			"o|output", "Output file. If unspecified, output is printed to stdout.", &args.output,
			"f|format", "Output format (either html or gemtext).", &args.fmt,
			"d|dir", "Treat input and output as directories. This will convert all *.md files in one directory and output them to other one, creating directories as needed.", &args.dir,
			"v|verbose", "Print out extra info.", &args.verbose,
			"t|template", "Specifies a template file. The string '$text$' will be replaced by the resulting text.", &args.template_
		);
		if(res.helpWanted || empty) {
			defaultGetoptPrinter("Options:", res.options);
			return 0;
		}
	} catch(GetOptException e) {
		writeln(e.msg);
		return 1;
	} catch(ConvException e) {
		writeln(e.msg);
		return 1;
	}
	if(args.input == "") {
		writeln("No input specified.");
		return 2;
	}
	if(args.fmt == Format.none) {
		writeln("No format specified.");
		return 2;
	}
	string template_ = "";
	if(args.template_ != "") {
		try {
			template_ = readText(args.template_);
		} catch(FileException e) {
			writeln("Could not find template file: ", e.msg);
		}
	}
	try {
		auto conv = Converter.format(args.fmt);
		if(!args.dir) {
			string text = readText(args.input);
			auto md = new Parser(text).parse();
			string output = conv.convert(md, template_);
			if(args.output == "") {
				writeln(output);
				return 0;
			}
			std.file.write(setExtension(args.output, conv.extension), output);
			if(args.verbose)
				writefln("Output written to %s.", args.output);
			return 0;
		}
		if(args.output == "") {
			writeln("No output directory specified.");
			return 2;
		}
		string path = buildNormalizedPath(args.input);
		string outpath = buildNormalizedPath(args.output);
		foreach(string file; dirEntries(path, SpanMode.breadth)) {
			if(file.extension != ".md")
				continue;
			string outname = buildPath(outpath, file[path.length+1..$].setExtension(conv.extension));
			mkdirRecurse(buildPath(outpath, outname.dirName));
			string text = readText(file);
			string output = conv.convert(new Parser(text).parse(), template_);
			std.file.write(outname, output);
			if(args.verbose)
				writeln("Converted "~file~" to "~outname);
		}
	} catch(FileException e) {
		writeln(e.msg);
		return 3;
	}
	return 0;
}
