/// Parses markdown and returns a tree.
module parser;

import std.typecons, std.sumtype, std.algorithm, std.array, std.format, std.string;

/// Type of a tree
enum TreeType {
	top, line, text, italic, bold, link, header, list, code, codeBlock
}

/// A markdown link
struct Link {
	string label, link;
}

/// A markdown header
struct Header {
	uint size;
	Tree[] children;
}

/// A markdown code block
struct CodeBlock {
	string code, lang;
}

// A markdown AST
struct Tree {
	TreeType type; /// Type
	union {
		string text; /// Text
		Tree[] children; /// Children (for top)
		Link link; /// Link
		Header header; /// Header
		CodeBlock codeBlock; /// Code Block
	}
	
	/// Create text tree
	static Tree mktext(string txt, TreeType type = TreeType.text) {
		Tree t = Tree(type);
		t.text = txt;
		return t;
	}
	
	/// Create top tree
	static Tree mktop(Tree[] children) {
		Tree t = Tree(TreeType.top);
		t.children = children;
		return t;
	}
	
	/// Create link
	static Tree mklink(string label, string link) {
		Tree t = Tree(TreeType.link);
		t.link = Link(label, link);
		return t;
	}

	/// Create header
	static Tree mkheader(uint size, Tree[] children) {
		Tree t = Tree(TreeType.header);
		t.header = Header(size, children);
		return t;
	}

	/// Create line
	static Tree mkline(Tree[] children) {
		Tree t = Tree(TreeType.line);
		t.children = children;
		return t;
	}
	
	/// Create list
	static Tree mklist(Tree[] children) {
		Tree t = Tree(TreeType.list);
		t.children = children;
		return t;
	}
	
	/// Create code block
	static Tree mkcodeblock(string code, string lang) {
		Tree t = Tree(TreeType.codeBlock);
		t.codeBlock = CodeBlock(code, lang);
		return t;
	}
	
	/// Convert a tag to a human-readable string
	string toString() {
		final switch(type) {
			case TreeType.top:
				return children.map!(c => c.toString()).join("\n");
			case TreeType.text:
				return "Text("~text~")";
			case TreeType.italic:
				return "Italic("~text~")";
			case TreeType.bold:
				return "Bold("~text~")";
			case TreeType.link:
				return "%s".format(link);
			case TreeType.header:
				return "%s".format(header);
			case TreeType.line:
				return "Line("~children.map!(c => c.toString()).join(" ")~")";
			case TreeType.list:
				return "List("~children.map!(c => c.toString()).join(" ")~")";
			case TreeType.code:
				return "Code("~text~")";
			case TreeType.codeBlock:
				return "%s".format(codeBlock);
		}
	}
}

/// Safely gets an element of an array
private Nullable!T safeGet(T)(T[] s, size_t idx) {
	return idx >= s.length ? typeof(return).init : typeof(return)(s[idx]);
}

/// A class that parses markdown
class Parser {
	string text; /// Text that is currently being parsed
	size_t pos = 0; /// Position in text
	size_t startPos = 0; // Start position for current token

	/// Initializes with text being parsed
	this(string text) {
		this.text = text;
	}
	
	/// Get the current token and set startPos
	private string tk() {
		string ret = text[startPos..pos];
		startPos = pos+1;
		return ret;
	}

	/// Find the next occurence of a given pattern.
	/// Returns position if it was found, null if it wasn't.
	Nullable!size_t find(size_t start, string pattern, bool stopAtNewline = true) {
		size_t i = start;
		outer: while(i < text.length) {
			char c = text[i];
			switch(c) {
				case '\r':
				case '\n':
					if(stopAtNewline)
						return typeof(return).init;
					break;
				default:
					break;
			}
			for(size_t j = 0; j < pattern.length; j++) {
				if(text.safeGet(i+j) != pattern[j]) {
					i++;
					continue outer;
				}
			}
			return typeof(return)(i);
		}
		return typeof(return).init;
	}
	
	/// Parse
	Tree parse() {
		Tree[] ret;
		Tree[] line;
		void addText() {
			string token = tk;
			if(token != "")
				line ~= Tree.mktext(token);
		}
		struct LineInfo {
			bool start = true; // if we're at the start of a line
			bool list = false; // whether this is a list item
			uint header = 0u; /// Header count (0 if no header)
		}
		LineInfo li;
		void start() {
			if(li.start) {
				startPos = pos;
				li.start = false;
			}
		}
		while(pos < text.length) {
			switch(text[pos]) {
				// whitespace
				case ' ':
				case '\t':
					if(li.start) {
						startPos = pos;
					}
					break;
				case '\r':
					break; // windows :P
				case '\n':
					addText;
					if(li.header > 0)
						ret ~= Tree.mkheader(li.header, line);
					else if(li.list) {
						if(ret.length > 0 && ret[$-1].type == TreeType.list)
							ret[$-1].children ~= Tree.mkline(line);
						else
							ret ~= Tree.mklist([Tree.mkline(line)]);	
					}
					else
						ret ~= Tree.mkline(line);
					line = [];
					li = LineInfo.init;
					break;
				case '[': {
					start;
					auto idx = find(pos, "]");
					if(idx.isNull)
						break;
					string label = text[pos+1..idx.get];
					if(text.safeGet(idx.get+1) != '(')
						break;
					size_t lstart = idx.get+2;
					idx = find(idx.get+1, ")");
					if(idx.isNull)
						break;
					string link = text[lstart..idx.get];
					addText;
					line ~= Tree.mklink(label, link);
					pos = idx.get+1;
					startPos = pos;
					break;
				}
				case '\\': {
					start;
					auto c = text.safeGet(pos+1);
					if(c.isNull)
						break;
					addText;
					line ~= Tree.mktext(""~c.get);
					startPos = ++pos;
					startPos++;
					break;
				}
				case '#':
					if(li.start)
						li.header++;
					break;
				case '*': {
					if(li.start && text.safeGet(pos+1) == ' ') {
						li.list = true;
						break;
					}
					if(text.safeGet(pos+1) != '*') {
						auto other = find(pos+1, "*");
						if(other.isNull)
							break;
						addText;
						line ~= Tree.mktext(text[pos+1..other.get], TreeType.italic);
						pos = other.get+1;
						startPos = pos;
						break;
					}
					auto other = find(pos+1, "**");
					if(other.isNull)
						break;
					addText;
					line ~= Tree.mktext(text[pos+2..other.get], TreeType.bold);
					pos = other.get+2;
					startPos = pos;
					break;
				}
				case '`': {
					if(text.safeGet(pos+1) == '`' && text.safeGet(pos+2) == '`') {
						// large code block
						auto other = find(pos+3, "```", false);
						if(other.isNull)
							break;
						addText;
						string code = text[pos+3..other.get];
						auto newline = code.indexOf('\n');
						string lang = "";
						if(newline > 0) {
							lang = code[0..newline];
							code = code[newline+1..$];
						}
						line ~= Tree.mkcodeblock(code, lang);
						pos = other.get+3;
						startPos = pos;
						break;
					}
					auto other = find(pos+1, "`");
					if(other.isNull)
						break;
					addText;
					line ~= Tree.mktext(text[pos+1..other.get], TreeType.code);
					pos = other.get+1;
					startPos = pos;
					break;
				}
				default:
					start;
					break;
			}
			pos++;
		}
		return Tree.mktop(ret);
	}
}
